@@ -20,6 +20,25 @@ _RandType = Union[np.random.Generator, int]
 # module-wide numpy random number generator instance
 _GLOBAL_RNG = np.random.default_rng()
 
+# Constants
+_DEFAULT_HALF_LIFE_CALCULATION = "arithmetic"
+# ^ Numerically stable, no other method is consistently or significantly better.
+# It also has a natural, explainable biological significance
+_EXP_EPSILON = 1.0
+# ^ Shorthand for _EXPONENTIAL_EPSILON
+# It was set to 1.0 and not a real machine epsilon such as `np.finfo(float).eps`
+# because the machine epsilon can lead to numerical instability when calculating
+# the harmonic and geometric means.
+_EXP_HALF_LIFE_CALCULATION_MODES = {
+    "arithmetic": np.mean,
+    "harmonic": ss.hmean,
+    "geometric": ss.gmean,
+    "median": np.median,
+    "midpoint": lambda x: (np.max(x) + np.min(x)) / 2.0,
+}
+_VALID_DROUPOUT_MODES = ("uniform", "exponential")
+_DEFAULT_DROPOUT_MODE = "exponential"
+
 
 def set_module_rng_seed(seed: int) -> np.random.Generator:
     """Fix the module-wide random number generator
@@ -83,7 +102,9 @@ def __sim_zero_inf(
     lambda
     """
     rng = rng or _GLOBAL_RNG
-    rng = np.random.default_rng(rng) if isinstance(rng, int) else rng
+    rng = (
+        np.random.default_rng(rng) if not isinstance(rng, np.random.Generator) else rng
+    )
     if not ignore_deprecation:
         __err_message = ["Error: ", "ZeroInf genes cannot be directly sampled"]
         raise DeprecationWarning("".join(__err_message))
@@ -96,7 +117,9 @@ def _sim_unimodal(
     """Simulate the expression of unimodal genes using
     a normal (gaussian) distribution."""
     rng = rng or _GLOBAL_RNG
-    rng = np.random.default_rng(rng) if isinstance(rng, int) else rng
+    rng = (
+        np.random.default_rng(rng) if not isinstance(rng, np.random.Generator) else rng
+    )
     return rng.normal(loc=mean, scale=std_dev, size=size)
 
 
@@ -118,7 +141,9 @@ def _sim_bimodal(
     Weights should be a tuple (or eventually a list) containing
     """
     rng = rng or _GLOBAL_RNG
-    rng = np.random.default_rng(rng) if isinstance(rng, int) else rng
+    rng = (
+        np.random.default_rng(rng) if not isinstance(rng, np.random.Generator) else rng
+    )
     # Parameters of the mixture components
     norm_params = np.array([[mean1, std_dev], [mean2, std_dev]])
     # A stream of indices from which to choose the component
@@ -135,17 +160,87 @@ def _dropout_mask(
 ) -> np.ndarray:
     """Dropout mask to obtain the same dropout_rate as originally estimated"""
     rng = rng or _GLOBAL_RNG
-    rng = np.random.default_rng(rng) if isinstance(rng, int) else rng
+    rng = (
+        np.random.default_rng(rng) if not isinstance(rng, np.random.Generator) else rng
+    )
     return rng.choice(
         (0, 1), size=size, replace=True, p=(dropout_rate, 1.0 - dropout_rate)
     )
 
 
+def _estimate_exponential_parameter(
+    data: Union[np.ndarray, pd.Series],
+    mode: str = _DEFAULT_HALF_LIFE_CALCULATION,
+) -> float:
+    """Estimate \\beta_1 for the exponential decay model
+    of drop-out probabilities.
+
+    Probabilistic model:
+
+    $$
+        p = \\beta_0 * e^{\\beta_1 * (x - x_0)}
+    $$
+
+    Note: \\beta_0 can be easily calculated as follows:
+
+        \\beta_0 = 2.0 * criteria["DropOutRate"]
+
+    This is the actual computation performed in
+    `scboolseq.scBoolSeq.simulation_fit`
+
+    returns:
+    -------
+        \\beta_1
+    """
+    positive_entries = data > 0.0
+    positive_data = data[positive_entries]
+
+    _min_positive_value = np.min(positive_data)
+    x = positive_data - _min_positive_value + _EXP_EPSILON
+    _mu = _EXP_HALF_LIFE_CALCULATION_MODES[mode](x)
+
+    # b_0 is now calculated in scboolseq.scBoolSeq.simulation_fit
+    # b_0 = 2.0 * dropout_rate
+    b_1 = np.log(2.0) / _mu
+
+    return b_1
+
+
+def exponential_decay_dropout(
+    data: Union[np.ndarray, pd.Series],
+    b_0: float,
+    b_1: float,
+    rng: Optional[_RandType] = None,
+):
+    """Simulate dropout using an exponential decay probabilistic strategy"""
+    rng = rng or _GLOBAL_RNG
+    rng = (
+        np.random.default_rng(rng) if not isinstance(rng, np.random.Generator) else rng
+    )
+
+    decayed = data.copy()
+    positive_entries = data > 0.0
+    positive_data = data[positive_entries]
+
+    # TEST THE NECESSITY OF THE _EXP_EPSILON IN THIS CONTEXT
+    _min_positive_value = np.min(positive_data)
+    x = positive_data - _min_positive_value  # + _EXP_EPSILON
+
+    dropout_probabilities = b_0 * np.exp(-b_1 * x)  # (x - _EXP_EPSILON))
+    dropout_probabilities[dropout_probabilities >= 1.0] = 0.95
+    dropout_events = rng.binomial(1, p=dropout_probabilities)
+    dropout_mask = 1.0 - dropout_events
+    decayed[positive_entries] *= dropout_mask
+
+    return decayed
+
+
 def sample_gene(
     criterion: pd.Series,
     n_samples: int,
     enforce_dropout_rate: bool = True,
     rng: Optional[_RandType] = None,
+    dropout_mode: str = _DEFAULT_DROPOUT_MODE,
 ) -> pd.Series:
     """Simulate the expression of a gene, using the information provided by
     the criteria dataframe of a profile_binr.scBoolSeq class.
@@ -162,8 +257,14 @@ def sample_gene(
     in order to preserve the dropout rate estimated whilst computing the criteria
     for the original expression dataset ?
     """
+    if dropout_mode not in _VALID_DROUPOUT_MODES:
+        raise ValueError(
+            f"Invalid dropout mode '{dropout_mode}'. Valid modes are: {_VALID_DROUPOUT_MODES}"
+        )
     rng = rng or _GLOBAL_RNG
-    rng = np.random.default_rng(rng) if isinstance(rng, int) else rng
+    rng = (
+        np.random.default_rng(rng) if not isinstance(rng, np.random.Generator) else rng
+    )
     _data: np.array
 
     if criterion["Category"] == "Discarded":
@@ -180,7 +281,17 @@ def sample_gene(
         natural_dor = sum(_negative_data) / len(_data)
         if enforce_dropout_rate and natural_dor < criterion["DropOutRate"]:
             correction_dor = criterion["DropOutRate"] - natural_dor
-            _data *= _dropout_mask(dropout_rate=correction_dor, size=n_samples, rng=rng)
+            if dropout_mode == "uniform":
+                _data *= _dropout_mask(
+                    dropout_rate=correction_dor, size=n_samples, rng=rng
+                )
+            else:
+                _data = exponential_decay_dropout(
+                    data=_data,
+                    b_0=criterion["beta_0"],
+                    b_1=criterion["beta_1"],
+                    rng=rng,
+                )
     elif criterion["Category"] == "Bimodal":
         _data = _sim_bimodal(
             mean1=criterion["gaussian_mean1"],
@@ -195,9 +306,21 @@ def sample_gene(
         natural_dor = sum(_negative_data) / len(_data)
         if enforce_dropout_rate and natural_dor < criterion["DropOutRate"]:
             correction_dor = criterion["DropOutRate"] - natural_dor
-            _data *= _dropout_mask(dropout_rate=correction_dor, size=n_samples, rng=rng)
+            if dropout_mode == "uniform":
+                _data *= _dropout_mask(
+                    dropout_rate=correction_dor, size=n_samples, rng=rng
+                )
+            else:
+                _data = exponential_decay_dropout(
+                    data=_data,
+                    b_0=criterion["beta_0"],
+                    b_1=criterion["beta_1"],
+                    rng=rng,
+                )
     elif criterion["Category"] == "ZeroInf":
-        _data = __sim_zero_inf(_lambda=criterion["lambda"], size=n_samples, rng=rng)
+        __err_message = ["Error: ", "ZeroInf genes cannot be directly sampled"]
+        raise DeprecationWarning("".join(__err_message))
+        # _data = __sim_zero_inf(_lambda=criterion["lambda"], size=n_samples, rng=rng)
     else:
         raise ValueError(f"Unknown category `{criterion['Category']}`, aborting")
 
@@ -209,10 +332,13 @@ def _sample_sequential(
     n_samples: int,
     enforce_dropout_rate: bool = True,
     rng: Optional[_RandType] = None,
+    dropout_mode: str = _DEFAULT_DROPOUT_MODE,
 ) -> pd.DataFrame:
     """Simulate samples from a criteria dataframe, sequentially"""
     rng = rng or _GLOBAL_RNG
-    rng = np.random.default_rng(rng) if isinstance(rng, int) else rng
+    rng = (
+        np.random.default_rng(rng) if not isinstance(rng, np.random.Generator) else rng
+    )
     return (
         criteria.apply(
             lambda y: sample_gene(
@@ -220,6 +346,7 @@ def _sample_sequential(
                 n_samples=n_samples,
                 enforce_dropout_rate=enforce_dropout_rate,
                 rng=rng,
+                dropout_mode=dropout_mode,
             ),
             axis=1,
         ).T.dropna(how="all", axis=1),
@@ -231,6 +358,7 @@ def sample_from_criteria(
     n_samples: int,
     enforce_dropout_rate: bool = True,
     n_threads: int = mp.cpu_count(),
+    dropout_mode: str = _DEFAULT_DROPOUT_MODE,
 ) -> pd.DataFrame:
     """
     Create a new expression dataframe from a criteria dataframe using multithreading
@@ -242,6 +370,7 @@ def sample_from_criteria(
         _sample_sequential,
         n_samples=n_samples,
         enforce_dropout_rate=enforce_dropout_rate,
+        dropout_mode=dropout_mode,
     )
 
     _df_splitted_ls: List[pd.DataFrame] = np.array_split(criteria, n_threads)
@@ -251,23 +380,41 @@ def sample_from_criteria(
     return pd.concat(ret_list, axis=1)
 
 
-def random_nan_binariser(binary_df: pd.DataFrame, rng: Optional[_RandType] = None):
+def random_nan_binarizer(
+    binary_df: pd.DataFrame,
+    probs: Tuple[float] = (0.5, 0.5),
+    rng: Optional[_RandType] = None,
+):
     """Assuming df is a binary matrix produced by calling
     profile_binr.scBoolSeq.binarise() this function should
     randomly resolve all NaN entries to either 1 or 0
 
-    This function operates on a deep copy of the df argument
+    first probability is the probability of resolving
+    to one and the second is the probability of resolving
+    to zero.
+    probs = (p(nan->1), p(nan->0))
+
+    This function operates on a deep copy of the df argument.
     """
     rng = rng or _GLOBAL_RNG
-    rng = np.random.default_rng(rng) if isinstance(rng, int) else rng
+    rng = (
+        np.random.default_rng(rng) if not isinstance(rng, np.random.Generator) else rng
+    )
     binary_df = binary_df.copy(deep=True)
-    for gene in binary_df.columns:
-        mask = binary_df[gene].isna()
-        binary_df.loc[mask, gene] = rng.choice((1.0, 0.0), mask.sum())
+    _na_mask = binary_df.isna()
+    _size = _na_mask.shape[0] * _na_mask.shape[1]
+    binary_df[_na_mask] = rng.choice((1.0, 0.0), size=_size, p=probs).reshape(
+        _na_mask.shape
+    )
 
     return binary_df
 
 
+def random_nan_binariser(*args, **kwargs):
+    """alias for scboolseq.simulation.random_nan_binarizer"""
+    return random_nan_binarizer(*args, **kwargs)
+
+
 def simulate_unimodal_distribution(
     value: float,
     mean: float,
@@ -290,7 +437,9 @@ def simulate_unimodal_distribution(
                         half normal distribution.
     """
     rng = rng or _GLOBAL_RNG
-    rng = np.random.default_rng(rng) if isinstance(rng, int) else rng
+    rng = (
+        np.random.default_rng(rng) if not isinstance(rng, np.random.Generator) else rng
+    )
     # random variate right hand side
     _rv_rhs = ss.halfnorm.rvs(loc=mean, scale=std_dev, size=size, random_state=rng)
     # random variate left hand side
@@ -302,6 +451,7 @@ def simulate_bimodal_gene(
     binary_gene: pd.Series,
     criterion: pd.Series,
     rng: Optional[_RandType] = None,
+    dropout_mode: str = _DEFAULT_DROPOUT_MODE,
 ) -> pd.Series:
     """
     Simulate bimodal gene expression, by sampling the appropriate
@@ -321,8 +471,11 @@ def simulate_bimodal_gene(
           the Boolean states.
     """
     rng = rng or _GLOBAL_RNG
-    rng = np.random.default_rng(rng) if isinstance(rng, int) else rng
-    assert binary_gene.name == criterion.name, "Criterion and gene mismatch"
+    rng = (
+        np.random.default_rng(rng) if not isinstance(rng, np.random.Generator) else rng
+    )
+    if binary_gene.name != criterion.name:
+        raise ValueError("Criterion and gene mismatch")
     simulated_normalised_expression = pd.Series(
         np.nan, index=binary_gene.index, name=criterion.name, dtype=float
     )
@@ -354,13 +507,19 @@ def simulate_bimodal_gene(
         # check how many values do we need to put to zero
         _correction_dor = criterion["DropOutRate"] - natural_dor
 
-        # calculate a correction mask
-        # (vector of 1 and 0 at random indices, according to the correction DOR)
-        simulated_normalised_expression *= _dropout_mask(
-            dropout_rate=_correction_dor,
-            size=len(simulated_normalised_expression),
-            rng=rng,
-        )
+        if dropout_mode == "uniform":
+            simulated_normalised_expression *= _dropout_mask(
+                dropout_rate=_correction_dor,
+                size=len(simulated_normalised_expression),
+                rng=rng,
+            )
+        else:
+            simulated_normalised_expression = exponential_decay_dropout(
+                data=simulated_normalised_expression,
+                b_0=criterion["beta_0"],
+                b_1=criterion["beta_1"],
+                rng=rng,
+            )
 
     return simulated_normalised_expression
 
@@ -369,6 +528,7 @@ def simulate_unimodal_gene(
     binary_gene: pd.Series,
     criterion: pd.Series,
     rng: Optional[_RandType] = None,
+    dropout_mode: str = _DEFAULT_DROPOUT_MODE,
 ) -> pd.Series:
     """
     Simulate unimodal gene expression from Boolean states by sampling
@@ -387,8 +547,11 @@ def simulate_unimodal_gene(
           the Boolean states.
     """
     rng = rng or _GLOBAL_RNG
-    rng = np.random.default_rng(rng) if isinstance(rng, int) else rng
-    assert binary_gene.name == criterion.name, "Criterion and gene mismatch"
+    rng = (
+        np.random.default_rng(rng) if not isinstance(rng, np.random.Generator) else rng
+    )
+    if binary_gene.name != criterion.name:
+        raise ValueError("Criterion and gene mismatch")
     simulated_normalised_expression = pd.Series(
         0.0, index=binary_gene.index, name=criterion.name, dtype=float
     )
@@ -420,13 +583,20 @@ def simulate_unimodal_gene(
 
     if natural_dor < criterion["DropOutRate"]:
         _correction_dor = criterion["DropOutRate"] - natural_dor
-        # calculate a correction mask (vector of 1 and 0 at random indices,
-        # according to the correction DOR)
-        simulated_normalised_expression *= _dropout_mask(
-            dropout_rate=_correction_dor,
-            size=len(simulated_normalised_expression),
-            rng=rng,
-        )
+
+        if dropout_mode == "uniform":
+            simulated_normalised_expression *= _dropout_mask(
+                dropout_rate=_correction_dor,
+                size=len(simulated_normalised_expression),
+                rng=rng,
+            )
+        else:
+            simulated_normalised_expression = exponential_decay_dropout(
+                data=simulated_normalised_expression,
+                b_0=criterion["beta_1"],
+                b_1=criterion["beta_1"],
+                rng=rng,
+            )
 
     return simulated_normalised_expression
 
@@ -435,6 +605,7 @@ def simulate_gene(
     binary_gene: pd.Series,
     criterion: pd.Series,
     rng: Optional[_RandType] = None,
+    dropout_mode: str = _DEFAULT_DROPOUT_MODE,
 ) -> pd.Series:
     """Simulate the expression of a gene, using the information provided by
     the criteria dataframe of a profile_binr.scBoolSeq class.
@@ -452,16 +623,22 @@ def simulate_gene(
     for the original expression dataset ?
     """
     rng = rng or _GLOBAL_RNG
-    rng = np.random.default_rng(rng) if isinstance(rng, int) else rng
+    rng = (
+        np.random.default_rng(rng) if not isinstance(rng, np.random.Generator) else rng
+    )
 
     if criterion["Category"] == "Discarded":
         return pd.Series(
             np.nan, index=binary_gene.index, name=criterion.name, dtype=float
         )
     elif criterion["Category"] == "Unimodal":
-        return simulate_unimodal_gene(binary_gene, criterion, rng=rng)
+        return simulate_unimodal_gene(
+            binary_gene, criterion, rng=rng, dropout_mode=dropout_mode
+        )
     elif criterion["Category"] == "Bimodal":
-        return simulate_bimodal_gene(binary_gene, criterion, rng=rng)
+        return simulate_bimodal_gene(
+            binary_gene, criterion, rng=rng, dropout_mode=dropout_mode
+        )
     else:
         raise ValueError(f"Unknown category `{criterion['Category']}`, aborting")
 
@@ -470,12 +647,17 @@ def _simulate_subset(
     binary_df: pd.DataFrame,
     simulation_criteria: pd.DataFrame,
     rng: Optional[_RandType] = None,
+    dropout_mode: str = _DEFAULT_DROPOUT_MODE,
 ) -> pd.Series:
     """helper function, wrapper for apply"""
     rng = rng or _GLOBAL_RNG
-    rng = np.random.default_rng(rng) if isinstance(rng, int) else rng
+    rng = (
+        np.random.default_rng(rng) if not isinstance(rng, np.random.Generator) else rng
+    )
     return binary_df.apply(
-        lambda x: simulate_gene(x, simulation_criteria.loc[x.name, :], rng=rng)
+        lambda x: simulate_gene(
+            x, simulation_criteria.loc[x.name, :], rng=rng, dropout_mode=dropout_mode
+        )
     )
 
 
@@ -485,6 +667,7 @@ def biased_simulation_from_binary_state(
     n_threads: Optional[int] = None,
     n_samples: Optional[int] = None,
     seed: Optional[int] = None,
+    dropout_mode: str = _DEFAULT_DROPOUT_MODE,
 ) -> pd.DataFrame:
     """n_threads defaults to multiprocessing.cpu_count()"""
 
@@ -505,18 +688,22 @@ def biased_simulation_from_binary_state(
     _seed_sequence_generator = np.random.SeedSequence(seed)
     child_seeds = _seed_sequence_generator.spawn(n_threads)
     streams = [np.random.default_rng(s) for s in child_seeds]
+    modes = len(child_seeds) * [dropout_mode]
 
     if n_threads > 1:
         _criteria_ls = np.array_split(simulation_criteria, n_threads)
         _binary_ls = np.array_split(binary_df, n_threads, axis=1)
         with mp.Pool(n_threads) as pool:
             ret_list = pool.starmap(
-                _simulate_subset, zip(_binary_ls, _criteria_ls, streams)
+                _simulate_subset, zip(_binary_ls, _criteria_ls, streams, modes)
             )
         return pd.concat(ret_list, axis=1)
     else:
         return _simulate_subset(
-            binary_df=binary_df, simulation_criteria=simulation_criteria, rng=streams[0]
+            binary_df=binary_df,
+            simulation_criteria=simulation_criteria,
+            rng=streams[0],
+            dropout_mode=dropout_mode,
         )
 
 
